/* infocon - identify blocks of high information content in an alignment, as
*  described by:
*
*  N. Stojanovic, L. Florea, C. Riemer, D. Gumucio, J. Slightom, M. Goodman,
*  W. Miller and R. Hardison (2000)
*  Comparison of five methods for finding conserved sequences in multiple
*  alignments of gene regulatory networks.
*  Nucleic Acids Research 27, 3899-3910.
*
*  Command syntax:
*	infocon refine-output-file [A=float] [R=int] [C=0-or-1]
*  where A and R set the anchor value and the minimum length.  C=1 (default 0)
*  requests that the consensus sequence be printed (instead of the first
*  sequence).
*/

#include "util.h"
#include "args.h"
#include "read_malign.h"
#include <math.h>

#define DEFAULT_ANCHOR 1.0
#define DEFAULT_MIN_RUN 6

/* Full runs. */
typedef struct run {
    int beg, end, SR, SL, R, L, Lower;
} Run, *Run_ptr;

/* consensus sequence; if two or more letters tie for the most frequent at a
*  given position, pick the first of them in the list A,C,G,T,-.
*/
char *consensus(char **A, int nrow, int ncol)
{
	int i, j, na, nc, ng, nt, ndash, c;
	char *consen = ckalloc(ncol*sizeof(char));

	for (j = 0; j < ncol; ++j) {
		na = nc = ng = nt = ndash = 0;
		for (i = 0; i < nrow; ++i) {
			if ((c = A[i][j]) == 'A')
				++na;
			else if (c == 'C')
				++nc;
			else if (c == 'G')
				++ng;
			else if (c == 'T')
				++nt;
			else if (c == '-')
				++ndash;
/*
			else
				fatalf("illegal alignment character: '%c'", c);
*/
		}
		if (na >= nc && na >= ng && na >= nt && na >= ndash)
			c = 'A';
		else if (nc >= na && nc >= ng && nc >= nt && nc >= ndash)
			c = 'C';
		else if (ng >= na && ng >= nc && ng >= nt && ng >= ndash)
			c = 'G';
		else if (nt >= na && nt >= nc && nt >= ng && nt >= ndash)
			c = 'T';
		else
			c = '-';
		consen[j] = c;
	}
	return consen;
}

int main(int argc, char **argv)
{
	char **A, **seqname, *consen;
	double fa, fc, fg, ft, fca, fcc, fcg, fct, denom, ratio, divide_by, S,
	  *score, anchor;
	int nrow, ncol, i, j, k, c, na, nc, ng, nt, top, L, b, e, min_run, C,
	  *col2pos;
	Run_ptr run;

	/* handle command-line argments */
	argv0 = "infocon";
	if (argc < 2)
		fatal("arg = refine-output-file [ACR=]");
	ckargs("ACR", argc, argv, 2);
	get_argval_pos('R', &min_run, DEFAULT_MIN_RUN);
	get_argval_nonneg('C', &C, 0);
	if (!get_fargval((int)'A', &anchor))
		anchor = DEFAULT_ANCHOR;
	
	/* get the alignment */
	A = read_malign(argv[1], &nrow, &ncol, &seqname);
	
	/* for conversion to base-2 logs and dividing by number of rows: */
	divide_by = log((double)2) * (double)nrow;

	/* determine frequencies of A, C, G and T in the entire alignment */
	na = nc = ng = nt = 0;
	for (j = 0; j < ncol; ++j)
		for (i = 0; i < nrow; ++i)
			if ((c = A[i][j]) == 'A')
				++na;
			else if (c == 'C')
				++nc;
			else if (c == 'G')
				++ng;
			else if (c == 'T')
				++nt;
	denom = (double)(na+nc+ng+nt);
	fa = (double)na/denom;
	fc = (double)nc/denom;
	fg = (double)ng/denom;
	ft = (double)nt/denom;

	/* find information content minus anchor of each alignment column */
	score = ckalloc(ncol*sizeof(double));
	for (j = 0; j < ncol; ++j) {
		na = nc = ng = nt = 0;
		for (i = 0; i < nrow; ++i)
			if ((c = A[i][j]) == 'A')
				++na;
			else if (c == 'C')
				++nc;
			else if (c == 'G')
				++ng;
			else if (c == 'T')
				++nt;
			else
				break;
		/* consider it minus infinity if column contains a '-' */
		if (i < nrow) {
			score[j] =  -1000.0;
			continue;
		}
		denom = (double)(na+nc+ng+nt);
		fca = (double)na/(denom*fa);
		fcc = (double)nc/(denom*fc);
		fcg = (double)ng/(denom*fg);
		fct = (double)nt/(denom*ft);
		ratio = 1.0;
		for (k = 0; k < na; ++k)
			ratio *= fca;
		for (k = 0; k < nc; ++k)
			ratio *= fcc;
		for (k = 0; k < ng; ++k)
			ratio *= fcg;
		for (k = 0; k < nt; ++k)
			ratio *= fct;
		score[j] = (log(ratio)/divide_by) - anchor;
	}

	/* find consensus sequence, giving most common letter at each column */
	consen = consensus(A, nrow, ncol);

	/* for each column, find the corresponding position in sequence 1 */
	col2pos = position_in_seq1(A, ncol);

	run = ckalloc((ncol+1)*sizeof(Run));

	/* Find the "full runs", i.e., segments of the alignment with higher
	*  total score (information content minus anchor value) that and of
	*  its sub-parts, and where any containing segment has lower score.
	*  The algorithm is described by:
	*  X. Huang, P. Pevzner and W. Miller (1994)
	*  Parametric recomputing in alignment graphs.  Combinatorial Pattern
	*  Matching.  (Lecture Notes in Computer Science 807; Springer-Verlag)
	*  pp. 87-101, especially Fig. 6.
	*/
	S = 0.0;
	top = 0;
	for (j = 0; j < ncol; ++j) {
		S += score[j];
		if (score[j] >= 0) {
			/* form a one-column alignment */
			top++;
			run[top].L = run[top].R = j;
			run[top].SL = S - score[j];
			run[top].SR = S;
			run[top].beg = run[top].end = j;
			L = top-1;
			while (L > 0 && run[L].SL > run[top].SL) 
				L = run[L].Lower;
			run[top].Lower = L;
			/* merge alignments */
			while (L > 0 && run[L].SR <= run[top].SR) {
		    		run[L].R = run[top].R;
				run[L].SR = run[top].SR;
	      			run[L].end = run[top].end;
	      			top = L;
				L = run[top].Lower;
			}
		}
	}
	for (i = 1; i <= top; i++)
		if ((run[i].end - run[i].beg + 1) >= min_run) {
			b = run[i].beg;
			e = run[i].end;
			printf("%d %d  ", col2pos[b], col2pos[e]);
			for (j = b; j <= e; ++j)
				putchar(C == 1 ? consen[j] : A[0][j]);
			putchar('\n');
		}

	return 0;
}
