/*
* pat -- search a multiple alignment for matches to a string of (possibly
*  ambiguous) *	nucleotides.
*
* The command
*	pat WGATAR multiple-alignment-file
* prints locations in the first sequence where WGATAR matches each row of the
* alignment contained in the * named file.  (W = A or T, and R = A or G).
* The command
*	pat -c WGATAR multiple-alignment-file
* in addition prints locations where the reverse complement of WGATAR
* matches the alignment.  Finally,
*	pat WGATAR multiple-alignment-file 2
* allows each row to have up to two mismatches with WGATAR.
*
* The nucleotide ambiguity codes are:
* B = C, G or T
* D = A, G or T
* H = A, C or T
* K = G or T
* M = A or C
* N = A, C, G or T
* R = A or G (purines)
* S = C or G
* V = A, C or G
* W = A or T
* X = A, C, G or T
* Y = C or T (pyrimidines)
*/

#include "util.h"
#include "read_malign.h"

typedef unsigned char uchar;

/*             ABCDEFGHIJKLMNOPQRSTUVWXYZ	*/
char comp[] = "TVGH  CD  M KN   YSA BWXR ";

char *matches[128];

void init() {
	matches['A'] = copy_string("A");
	matches['B'] = copy_string("BCGKSTY");
	matches['C'] = copy_string("C");
	matches['D'] = copy_string("ADGKRTW");
	matches['G'] = copy_string("G");
	matches['H'] = copy_string("ACHMTWY");
	matches['K'] = copy_string("GKT");
	matches['M'] = copy_string("ACM");
	matches['N'] = copy_string("ABCDGHKMNRSTVWXY");
	matches['R'] = copy_string("AGR");
	matches['S'] = copy_string("CGS");
	matches['T'] = copy_string("T");
	matches['V'] = copy_string("ACGMRSV");
	matches['W'] = copy_string("ATW");
	matches['X'] = copy_string("ABCDGHKMNRSTVWXY");
	matches['Y'] = copy_string("CTY");
	matches['-'] = copy_string("-");
}

int main(int argc, char *argv[])
{
	char *P, *C, **seqname, **A, *p;
	int M, nrow, ncol, i, j, k, do_compl = 0, miss = 0, m, *col2pos;

	argv0 = "pat";
	if (argc > 1 && same_string(argv[1], "-c")) {
		++argv;
		--argc;
		do_compl = 1;
	}
	if (argc == 4) {
		if ((miss = atoi(argv[3])) <= 0)
			fatal("number of mismatches must be positive");
	} else if (argc != 3)
		fatal("arguments are [-c] pattern alignment [mismatches]");
	P = argv[1];
	M = strlen(P);
	if (do_compl) {
		C = ckalloc((M+1) * sizeof(char));
		for (i = 0; i < M; ++i)
			C[i] = comp[P[M-1-i]-'A'];
		C[M] = '\0';
	}
	A = read_malign(argv[2], &nrow, &ncol, &seqname);
	/* determine mapping from alignment columns to positions in seq 1 */
	col2pos = position_in_seq1(A, ncol);
	init();
	printf("matches to \"%s\" ",argv[1]);
	if (do_compl)
		printf("or its reverse complement\n\t");
	if (miss)
		printf("(allowing %d mismatch%s) ", miss, miss==1?"":"es");
	printf("in alignment file \"%s\"\n", argv[2]);
	for (j = 0; j <= ncol - M; ++j) {
		for (i = 0; i < nrow; ++i) {
			for (p = P, m = k = 0; k < M; ++k, ++p)
				if (!strchr(matches[(uchar)A[i][j+k]], *p)
				   && ++m > miss)
					break;
			if (k < M)
				break;
		}
		if (i < nrow && do_compl) {
		    for (i = 0; i < nrow; ++i) {
			for (p = C, m = k = 0; k < M; ++k,--p)
				if (!strchr(matches[(uchar)A[i][j+k]], *p)
				    && +m > miss)
					break;
			if (k < M)
				break;
		    }
		}
	    	if (i == nrow) {
		    	printf("%d ", col2pos[j]);
		    	for (p = A[0]+j, k = 0; k < M; ++k, ++p)
			    	putchar(*p);
		    	putchar('\n');
	    	}
	}
	 

	return 0;
}
