/* lav.c - read an "lav" file */

#include <errno.h>

#include "util.h"
#include "lav.h"

static const char rcsid[]=
"$Id: lav.c,v 1.8 2003/01/22 22:18:19 schwartz Exp $";

enum {
	BUF_SIZE=16*1024, MAX_LABELS=30
};

static const char Align[] =
"#:lav\n\
d {\n\
  \"%s\"\n\
}\n\
s {\n\
  \"fakeseq1\" %d %d\n\
  \"fakeseq2\" %d %d\n\
}\n\
h {\n\
   \">fakeheader1\"\n\
   \">fakeheader2\"\n\
}\n\
a {\n\
  s 0\n\
  b 1 2\n\
  e 1 2\n";

void align_header(int len1, int len2)
{
	align_header4(1, len1, 1, len2);
}

void align_header4(int L1, int R1, int L2, int R2)
{
	printf(Align, (argv0?argv0:"PipMaker"), L1, R1, L2, R2);
}

static const char Tag[] =
"/x-axis-label () def\n\
/y-axis-label () def\n\n\
0 setrow\n";

FILE *alignopen(char *alignfile)
{
	FILE *fp;
	char buf[BUF_SIZE];

	fp = ckopen(alignfile, "r");

	if (fgets(buf, BUF_SIZE, fp) == NULL)
		fatal("Empty alignment file.");
	if (strcmp(buf, "#:lav\n"))
		fatal("Not an lav file.");
	return fp;
}

/* seq1bounds - return length and endpoints of the first sequence, after checking for errors */
int seq1bounds(char *alignfile, int *l, int *r)
{
	char buf[BUF_SIZE], name[BUF_SIZE];
	int P1, P1old, P2, P2old;
	FILE *fp;

	fp = alignopen(alignfile);

	P1old = 0;
	P2old = 0;
	buf[0] = '\0';
	while (lav_seqname(1, fp, buf, BUF_SIZE, name, &P1, &P2)) {
		if (P1old == 0)
			P1old = P1;
		else if (P1old != P1)
			fatalf("start of seq1 is not always %d.", P1);

		if (P2old == 0)
			P2old = P2;
		else if (P2old != P2)
			fatalf("end of seq1 is not always %d.", P2);
	}
	fclose(fp);
	if (l) *l = P1;
	if (r) *r = P2;
	return P2-P1+1;
}

/* seq1len - return length of the first sequence, after checking for errors */
int seq1len(char *alignfile)
{
	return seq1bounds(alignfile, 0, 0);
}

int seq2len(char *alignfile)
{
	char buf[BUF_SIZE], name[BUF_SIZE];
	int P1, L2, L2tot;
	FILE *fp;
	int rc, nth, prevth;

	fp = alignopen(alignfile);

	rc = nth = prevth = -1;
	L2tot = 0;
	buf[0] = '\0';
	while (lav_seqname4(2, fp, buf, BUF_SIZE, name, &P1, &L2, &rc, &nth)) { 
		if (nth != prevth)
			L2tot += (L2-P1+1);
		prevth = nth;
	}
	fclose(fp);
	return L2tot;
}

/* maybe some day fix for filenames ending with '-' */
int isrevcomp(char *oldname, char *name)
{
	int len = strlen(name);

	return (len > 0 && (name[len-1] == '-' ||
	                    (name[len-1] == '"' && name[len-2] == '-')));
}

/* dotclash - compute minimum separation of White Arrow labels in dotplot */
int dotclash(char *alignfile, int nseq)
{
	int L1, L2;

	L1 = seq1len(alignfile);
	if (nseq == 1)
		return L1/MAX_LABELS;
	L2 = seq2len(alignfile);
	return (MAX(L1,L2))/MAX_LABELS;
}


int lav_scan_s(char *line, char *name, int *lo, int *hi, int *rc, int *nth)
{
    return sscanf(line, " %1000s %d %d %d %d", name, lo, hi, rc, nth);
}

char *lav_seqname(int which, FILE *fp, char *buf, int bsize, char *name,
  int *from_addr, int *to_addr)
{
    int a;
    return lav_seqname4(which,fp,buf,bsize,name,from_addr,to_addr,&a,&a);
}

char *lav_seqname4(int which, FILE *fp, char *buf, int bsize, char *name,
  int *from_addr, int *to_addr, int *rc, int *nth)
{
	char buf2[BUF_SIZE], *p; 
	int n;

	while (strcmp(buf, "s {\n"))
		if (fgets(buf, bsize, fp) == NULL) {
			name[0] = '\0';
			return NULL;
		}
	fgets(buf, bsize, fp);	/* first sequence */
	if (which > 1)
		fgets(buf, bsize, fp); /* 2nd sequence */
	n = lav_scan_s(buf,buf2,from_addr,to_addr,rc,nth);
	if (n == 3) { rc = 0; nth = 0; }
	else if (n == 5) { /* ok */ }
	else
	    fatalf("%s\nCannot parse sequence name", buf);
	if ((p = strchr(buf2+1,'"')) != NULL)
		*p = '\0';
	strcpy(name, buf2+1);
	return name;
}


char *lav_contigname(FILE *fp, char *buf, int bsize, char *contig_name)
{
	char *s;

	while (strcmp(buf, "h {\n"))
		if (fgets(buf, bsize, fp) == NULL) {
			contig_name[0] = '\0';
			return NULL;
		}
	fgets(buf, bsize, fp);
	fgets(buf, bsize, fp);
	if ((s = strchr(buf, '"')) != NULL) {
		strcpy(contig_name, s+1);
		if ((s = strchr(contig_name, '"')) != NULL)
			*s = '\0';
	}
	return contig_name;
}


int same_prefix(char *line, char *prefix)
{
        if (line == prefix) return 1;
        if (line == 0) return 0;
        if (prefix == 0) return 0;
        return strncmp(line, prefix, strlen(prefix)) == 0;
}

// compare h{} lines, modulo reverse complement
int same_header_rc(char *a, char *b)
{
        char *ar = strstr(a, " (reverse complement)");
        char *br = strstr(b, " (reverse complement)");
        char *an = ar ? ar : strchr(a, 0);
        char *bn = br ? br : strchr(b, 0);
        while (a <= an && b <= bn && *a == *b) {
                ++a; ++b;
        }
        return a == an || b == bn;
}

int same_contig_again(char *line)
{
        static char *previous = 0;
        int same = previous && same_header_rc(previous,line);
        if (previous) free(previous);
        previous = copy_string(line);
        return same;
}

