/* popup.c --- generate pop-up messages to appear in a pip

   Syntax is "popup blastz-output-file".
   The idea is to completely cover the pip with messages, either about the
   local alignment where the mouse is pointing or the current gap.
*/

static const char rcsid[] = 
"$Id: popup.c,v 1.2 2002/07/23 18:06:56 schwartz Exp $";

#include "util.h"
#include "piplib.h"

enum { IGNORE=200	/* ignore tiny alignments IF NESTED */
     , MBLOCK=100000	/* maximum total number of blocks */
};

typedef struct bl {
	int b1, b2, e1, e2;
	char *contig_name;
} Block;

Block block[MBLOCK];
int nblock = 0;

void free_contig_names()
{
	int i;

	for (i = 0; i < nblock; ++i)
		if (block[i].contig_name)
			free(block[i].contig_name);
}

int get_len1(FILE *fp)	/* length of sequence 1 */
{
	char buf[500];
	int len;

	while (fgets(buf, 500, fp))
		if (strncmp(buf, "s {", 3) == 0) {
			if (fgets(buf, 500, fp) == NULL)
				return 0;
			if (sscanf(buf+2, "%*s %*d %d", &len) != 1)
				fatalf("len1: %s", buf);
			return len;
		}
	return 0;
}

int block_cmp(const void *a, const void *b)
{
	return ((const Block *)a)->b1 - ((const Block *)b)->b1;
}

void white(int len1)
{
	int i, nesting, len;

	qsort((char *) block, nblock, sizeof(Block), block_cmp);

	if (nblock == 0) {
		printf("1 %d (%d bp unmatched) White Block\n", len1, len1);
		return;
	}
	printf("1 %d (%d bp unmatched) White Block\n", block[0].b1-1,block[0].b1-1);
	for (nesting = i = 0; i < nblock; ++i) {
		len = block[i].e1 - block[i].b1;
		if (len > 0) {		/* start of local alignment */
			if (++nesting == 1 || len >= IGNORE) {
				printf("%d %d ", block[i].b1, block[i].e1);
				printf("(%d-%d matches %d-%d of %s) White Block\n",
					block[i].b1, block[i].e1, block[i].b2,
					block[i].e2, block[i].contig_name);
			}
		} else if (--nesting == 0 && i < nblock-1) {
			printf("%d %d (%d bp gap; ",
				block[i].e1+1, block[i+1].b1-1,
				block[i+1].b1-block[i].e1-1);
			if (same_string(block[i].contig_name,
					block[i+1].contig_name))
				printf("%d bp in %s)",
				    block[i+1].b2 - block[i].e2 - 1,
				    block[i].contig_name);
			else
				printf("between %s and %s)",
					block[i].contig_name,
					block[i+1].contig_name);
			printf(" White Block\n");
		}
	}
	if (nesting != 0)
		fatal("Impossible nesting.");
	printf("%d %d (%d bp unmatched) White Block\n",
	    block[nblock-1].e1+1, len1, len1 - block[nblock-1].e1);
}

int contig_len(FILE *fp, char *buf, int buf_size)
{
	int c_len;
	char *p;

	/* at top of LAV block; find contig length */
	while ((p = fgets(buf, buf_size, fp)) && strncmp(buf, "s {", 3))
		;
	if (p == NULL)
		fatal("contig: cannot happen");
	fgets(buf, buf_size, fp);	/* first seq */
	fgets(buf, buf_size, fp);
	if (sscanf(buf+2, "%*s %*d %d", &c_len) != 1)
		fatalf("len2: %s", buf);

	/* finish with the contig name in buf */
	while ((p = fgets(buf, 500, fp)) && strncmp(buf, "h {", 3))
		;
	if (p == NULL)
		fatal("contig: impossible");
	fgets(buf, buf_size, fp);	/* first seq */
	fgets(buf, buf_size, fp);

	return c_len;
}

static int is_space(int ch)
{
	return isascii(ch) && isspace(ch);
}

/* get alignments in an lav block */
int lavblock(FILE *fp)
{
	int c_len, b1, b2, e1, e2;
	char buf[500], name[500], *trim_name, *q, *rc;

	c_len = contig_len(fp, name, 500);
	rc = strstr(name, "(reverse complement)");
	trim_name = name;
	while (*trim_name == '"' || *trim_name == '>' || is_space(*trim_name))
		++trim_name;
	if (*trim_name == '\0')
		strcpy(trim_name, "Unnamed\""); /* fatalf("Cannot find contig name: %s", name); */
	q = trim_name+1;
	while (*q && !is_space(*q) && *q != '"')
		++q;
	*q = '\0';
	if (rc)
		strcat(trim_name,"-");

	quote_for_ps(trim_name);

	for ( ; ; ) {
		/* find next alignment or end of LAV block */
		while (fgets(buf, 500, fp)
		   && strncmp(buf, "#:lav", 5)
		   && strncmp(buf, "s {", 3)
		   && strncmp(buf, "a {", 3))
			;
		if (strncmp(buf, "a {", 3))
			break;
		fgets(buf, 500, fp);	/* score */
		fgets(buf, 500, fp);
		if (sscanf(buf, "  b %d %d", &b1, &b2) != 2)
			fatalf("no beginnings: %s", buf);
		fgets(buf, 500, fp);
		if (sscanf(buf, "  e %d %d", &e1, &e2) != 2)
			fatalf("no endings: %s", buf);
		block[nblock].b1 = b1;
		block[nblock].e1 = e1;
		block[nblock].b2 = b2;
		block[nblock].e2 = e2;
		block[nblock].contig_name = copy_string(trim_name);
		if (++nblock >= MBLOCK)
			fatal("Too many blocks.");
		/* 0-width block shows end of alignment */
		block[nblock].b1 = block[nblock].e1 = e1;
		block[nblock].b2 = block[nblock].e2 = e2;
		block[nblock].contig_name = copy_string(trim_name);
		if (++nblock >= MBLOCK)
			fatal("Too many blocks.");
	}

	return (strncmp(buf, "#:lav", 5) == 0);
}

int main(int argc, char **argv)
{
	char buf[500];
	FILE *fp;
	int len1;

	argv0 = "popup";
	if (argc != 2)
		fatal("arg = alignment_file");

	fp = ckopen(argv[1], "r");
	if (fgets(buf, 500, fp) == NULL || strncmp(buf, "#:lav", 5))
		fatal("Not a LAV alignment file.");
	len1 = get_len1(fp);
	rewind(fp);
	while (lavblock(fp))
		;
	fclose(fp);

	white(len1);

	free_contig_names();

	return 0;
}
