/* gene_tag.c -- make a gene+exon tag file from a simple "exons file" */

/* syntax:
*	gene_tag [-h] [-n] [-integer] exons_file
* where -h asks for a header file only, -n asks that exons not be numbered,
* and -integer (e.g., -7943) asks for debugging, assuming that the sequence
* length is the specified integer;
*/

#include "util.h"
#include "exons.h"
#include "piplib.h"
#include "genedir.h"

int main(int argc, char **argv)
{
	char *p, **gene_name, **exon_name;
	int i, j, k, n, first_exon, labeled, middle, nexons, incr, e, seq_len;
	int CDS_beg, CDS_end, count_exons = 1, *beg, *end, *dir;

	argv0 = "gene_tag";

	/* -h, -n and -integer  don't make sense together */
	if (argc > 1 && same_string(argv[1], "-h")) {
		if ((p = exon_header(argv[2])) != NULL) {
			quote_for_ps(p);
			printf("3 setrow\n  [[(%s) /Times-Roman]] Banner\n", p);
			free(p);
		}
		exit(0);
	}
	if (argc == 3 && (seq_len = atoi(argv[1]+1)) > 0) {
		nexons = get_exons(argv[2], &beg, &end, &dir, &gene_name,
		   &exon_name, 1);
		for (i = 0; i < nexons; ++i)
			if (end[i] > seq_len) {
			  fprintf(stderr, "Exon position is %d, ",end[i]);
			  fprintf(stderr, "whereas sequence length is %d.\n",
				seq_len);
			  fatal("error in sequence or exons file.");
			}
		free_exons(beg, end, dir, gene_name, exon_name, nexons);
		exit(0);
	}
	if (argc > 1 && same_string(argv[1], "-n"))
		count_exons = 0;
	else if (argc != 2)
		fatal("args = [-h] [-n] [-integer] exons_file");

	nexons = get_exons(argv[argc-1], &beg, &end, &dir, &gene_name,
	  &exon_name, 0);

	tag_header();

	n = 0;
	while (n < nexons) {
		if (gene_name[n] == NULL)
			fatal("syntax error in exons file: Cannot find gene name.");
		quote_for_ps(gene_name[n]);
		printf("1 setrow\n");
		printf("%d %d [[(%s) /Times-Roman]] %s Gene\n",
		    beg[n], end[n], gene_name[n], genedirstr(dir[n]));
		if (++n == nexons)
			fatalf("gene %s has no CDS information.",
				gene_name[n-1]);
		CDS_beg = beg[n];
		CDS_end = end[n];
		for (first_exon = ++n; n < nexons && gene_name[n] == NULL; ++n)
			;
		if (dir[first_exon] == 1) {
			e  = first_exon;
			incr = +1;
		} else {
			e = n - 1;
			incr = -1;
		}
		printf("0 setrow\n");
		labeled = -1000;
		/* don't print exon number if too near previous label */
		for (k = 1; k <= n - first_exon; ++k, e += incr) {
			i = beg[e];
			j = end[e];
			middle = (i+j)/2;
			if (count_exons &&
			  ((exon_name[e] == NULL && k < 9
			    && abs(middle-labeled) > 200)
			      || abs(middle-labeled) > 500)) {
				if (exon_name[e] == NULL)
				    printf("%d %d [[(%d) /Times-Roman]] Xon\n",
					i, j, k);
				else {
				    quote_for_ps(exon_name[e]);
				    printf("%d %d [[(%s) /Times-Roman]] Xon\n",
					i, j, exon_name[e]);
				}
				labeled = middle;
			} else
				printf("%d %d Exon\n", i, j);
		}
		for (k = first_exon; k < n; ++k) {
			i = beg[k];
			j = end[k];
			if (i < CDS_beg)
				printf("%d %d UTR\n", i, MIN(j, CDS_beg-1));
			if (CDS_end > 0 && j > CDS_end)
				printf("%d %d UTR\n", MAX(i,CDS_end+1), j);
		}
	}
	free_exons(beg, end, dir, gene_name, exon_name, nexons);
	return 0;
}
